%ifndef MESMACRO_INC
%define MESMACRO_INC
; min, max inclusive
; due to an oversight in expression processing, this will only work as the first value of an expression, otherwise it will count as an illegal operation
%define EXPRAND(min, max) 32h, (min) & 0xff, (min) >> 8 & 0xff, (max) & 0xff, (max) >> 8 & 0xff
;
%define RESOLVE(=x) x
%define EXPCONST1(v) 01h, (v)&255
%define EXPCONST2(v) 02h, (v)&255, (v)>>8&255
%define EXPCONST3(v) 03h, (v)&255, (v)>>8&255, (v)>>16&255
%define EXPCONST4(v) 04h, (v)&255, (v)>>8&255, (v)>>16&255, (v)>>24&255
%define EXPCONSTX(v) EXPCONST%+RESOLVE(1+((v)>255)+((v)>65535)+((v>16777215)))(v)
; lhs, rhs
%define EXPMUL(params+) params, 20h
%define EXPDIV(params+) params, 21h
%define EXPREM(params+) params, 22h
%define EXPADD(params+) params, 23h
%define EXPSUB(params+) params, 24h
%define EXPAND(params+) params, 25h
%define EXPOR(params+) params, 26h
; lhs, rhs
%define EXPCMPEQ(params+) params, 28h
%define EXPCMPNE(params+) params, 29h
%define EXPCMPLT(params+) params, 2ah
%define EXPCMPLTE(params+) params, 2bh
%define EXPCMPGT(params+) params, 2ch
%define EXPCMPGTE(params+) params, 2dh
%define EXPCMPLAND(params+) params, 2eh
%define EXPCMPLOR(params+) params, 2fh
;
%define EXPTERM(v+) v, 00h
; format is TYPE_REGISTERBASE_REGWIDTH_INDEXWIDTH
; REGWIDTH is the size of the value returned from the calculated address
; type REG accesses the register at a hard-coded 16-bit offset from base, calculated as (base + offset)
; types REGIDX amd DATIDX have a hard-coded offset from register base, but also have an additional variable expression IDX for indexing
; REGIDX returns the value from (base + offset + index * indexwidth)
; DATIDX returns the value from (*(base + offset) + index * indexwidth)
%define REGIDX_0000_1_1 05h
%define REGIDX_1000_2_2 06h
%define REGIDX_1800_4_4 07h
; 
%define DATIDX_1a00_1_1 08h
%define DATIDX_1a00_2_2 09h
%define DATIDX_1a00_4_4 0ah
;
%define REG_0000_1_1 0bh
%define REG_1000_2_2 0ch
%define REG_1800_4_4 0dh
%define REG_1a10_1_32 0eh
%define REG_1b88_2_2 0fh
%define REG_1a00_2_2 10h
%define REG2_1000_2_2 11h
%define REG_1b20_2_2 12h
;
%define EXPREG(type, reg) type, (reg) & 255, (reg) >> 8 & 255
%define EXPINDEX(type, reg, idx+) type, (reg) & 255, (reg) >> 8 & 255, idx
;
%define EXPSTR(str+) 11h, str, 00h
%define REGTERM(reg+) reg, 00h
;
%define LIST(vals+) vals, 00h
%define EMPTYLIST 00h
; special case, if first byte of expression is 00h, value is 1
; this should not be used in expression/value lists or other 00h terminated contexts, since this will be misinterpreted, a constant expression with 1 should be used instead
%define EXP1 00h
; operations
;
;
%macro endofcode 0
  db 00h
%endmacro
;
; 0x30 - illegal
;
; ifindex, endlabel, conditionexpr
%macro ifop 3+
  db 31h
  dw %1
  dw %2
  db %3
%endmacro
; ifindex, endlabel
%macro elseop 2
  db 32h
  dw %1
  dw %2
%endmacro
; endlabel, conditionexpr
%macro startloopop 2+
  db 33h
  dw %1
  db %2
%endmacro
; endlabel, valexpr
%macro selectop 2+
  db 34h
  dw %1
  db %2
%endmacro
; endlabel, listofcasevalexprs
%macro caseop 2+
  db 35h
  dw %1
  db %2
%endmacro
;
; 0x36 - illegal
;
; [depthexpr]
%macro break 0-1+
  db 37h
  %if %0 > 0
    db %1
    db 00h
  %else
    db EXP1
  %endif
%endmacro
; continue
%macro continue 0
  db 38h
%endmacro
; continuelabel, farlabel, conditionexpr, listexpr(string name)
%macro startmenuop 3+
  db 39h
  dw %1
  dw %2
  db %3
%endmacro
; u16 menuitemindex, u16 minbytex, u16 miny, u16 maxbytex, u16 maxy, u16 cmdlistofs, u16 id
%macro setmenuitemdata 1+
  db 3ah, 01h
  db EXPTERM(EXPCONSTX(1))
  db %1
  db 00h
%endmacro
; u16 count
%macro setmenuitemcount 1+
  db 3ah, 01h
  db EXPTERM(EXPCONSTX(2))
  db %1
  db 00h
%endmacro
; reg idreg
%macro menusetidreg 1+
  db 3ah, 02h
  db %1
  db 00h
%endmacro
; reg lastcheckedidxreg
%macro menusetlastcheckedidxreg 1+
  db 3ah, 03h
  db %1
  db 00h
%endmacro
; u16 bytexpos, u16 ypos, u8 columnwidth, u8 maxrows, u16 columnbytespacing, u16 ymargin, u16 menuitembytespacing, u16 menuitemcolumnspacing, u8 menuitemcangooffscreen
%macro setmenuproperties 1+
  db 3ah, 04h
  db %1
  db 00h
%endmacro
; u8 copyswapdstpageidx, u16 dstbytex, u16 dsty, reg menubyteposx, reg menubyteposy
%macro setmenuproperties2 1+
  db 3ah, 05h
  db %1
  db 00h
%endmacro
; reg selectedidxreg
%macro menusetselectedidxreg 1+
  db 3ah, 06h
  db %1
  db 00h
%endmacro
; u8 amount
%macro menudecreaseselectedidx 1+
  db 3ah, 07h
  db %1
  db 00h
%endmacro
; u8 amount
%macro menuincreaseselectedidx 1+
  db 3ah, 08h
  db %1
  db 00h
%endmacro
; u16 menuitemindex
%macro menurunitemcmdlistbyidx 1+
  db 3ah, 09h
  db %1
  db 00h
%endmacro
; output: u16 menuitemcount, u8 menucolumns, u8 menurows, u16 menubyteposx, u16 menuposy, u16 menuoutermaxbytex, u16 menuoutermaxy
%macro menureadproperties 1+
  db 3ah, 0ah
  db EXPTERM(EXPCONSTX(0))
  db %1
  db 00h
%endmacro
; input: u8 menuitemindex ; output: u16 minbytex, u16 miny, u16 maxbytex, u16 maxy, u16 cmdlistoffset, u16 id
%macro menuitemreadproperties 1+
  db 3ah, 0ah
  db %1
  db 00h
%endmacro
; 
%macro parsemenu 0
  db 3ah, 0bh
  db 00h
%endmacro
; 
%macro menugetselectedid 0
  db 3ah, 0ch
  db 00h
%endmacro
; const int expr menuitem id
%macro startmenuitemop 1+
  db 3bh
  db %1
  db 00h
%endmacro
;
%macro endmenuitemop 0
  db 3ch
  db 4 dup (0ffh)
  db 00h
%endmacro
;
%macro mousedragmenu 0
  db 3dh
%endmacro
; u8 optype
%macro menudisplayimagecopyswap 1
  db 3eh
  db %1
  db 00h
%endmacro
; targetlabel
%macro goto 1+
  db 3fh
  dw %%endlabel
  dw %1
  db EXP1
  db EMPTYLIST
  %%endlabel:
%endmacro
; targetlabel, listexpr(string name)
%macro goto2 2+
  db 3fh
  dw %%endlabel
  dw %1
  db EXP1
  db %2
  %%endlabel:
%endmacro
; targetlabel, conditionexpr, listexpr(string name)
%macro goto3 2+
  db 3fh
  dw %%endlabel
  dw %1
  db %2
  %%endlabel:
%endmacro
; continuelabel, targetlabel, conditionexpr, listexpr(string name)
%macro goto4 3+
  db 3fh
  dw %1
  dw %2
  db %3
%endmacro
; targetlabel
%macro call 1
  db 40h
  dw %%endlabel
  dw %1
  db EXP1
  db 00h
  %%endlabel:
%endmacro
; targetlabel, conditionexpr
%macro call2 2+
  db 40h
  dw %%endlabel
  dw %1
  db %2
  db 00h
  %%endlabel:
%endmacro
; continuelabel, targetlabel, conditionexpr
%macro call3 3+
  db 40h
  dw %1
  dw %2
  db %3
  db 00h
%endmacro
; 
%macro ret 0
  db 41h
%endmacro
;
; 0x42 - illegal
;
%define SKIPREG 31h, 00h
%define FILLREG 30h
; reg dst, reg(type 0e)/expr src[, SKIPREG, reg(type 0e)/expr srcN]*[, FILLREG, reg endrange]
%macro setreg 1+
  db 43h
  db %1
  db 00h
%endmacro
; {reg dst}, [reg(type 0f)/expstring]
%macro setstring 2+
  db 44h
  db %1
  dw %%endlabel
  db %2
  %%endlabel:
%endmacro
; {reg dst}, reg src
%macro copystring 2+
  db 45h
  db %1
  db 0ffh
  db %2
  db 00h
%endmacro
; {reg dst}, byte stringtype, byte stringdata[]
%macro copystringconst 3+
  db 45h
  db %1
  db 0ffh
  db %2
  db %3
  db 00h
  db 00h
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax
%macro settextbounds 1+
  db 46h
  db %1
%endmacro
; u8 xbytepos, u16 ypos
%macro clamptextposition 1+
  db 47h
  db %1
%endmacro
; TODO
%macro saverestoretextbounds 1+
  db 48h
  db %1
%endmacro
; u16 framesperglyph
%macro settextspeed 1+
  db 49h
  db %1
%endmacro
; printable text
%macro drawtext 1+
  db 4ah
  db %1
  db 00h
%endmacro
; reg src
%macro drawtextfromregister 1+
  db 4bh
  db %1
  db 00h
%endmacro
; u32 number[, u8 stringlen]
%macro drawnumber 1+
  db 4ch
  db %1
%endmacro
; u8 paletteindices ; high nibble background palette index, low nibble foreground/text palette index
%macro setdrawpaletteindices 1+
  db 4dh
  db %1
%endmacro
; u8 paletteindex
%macro settextcolor 1+
  db 4eh
  db %1
%endmacro
; [u8 fillpaletteindex]
%macro cleartextbox 1+
  db 4fh
  db %1
%endmacro
; [u8 fillpaletteindex]
%macro showtextbox 1+
  db 50h
  db %1
%endmacro
; u8 index, u16 color
%macro setpalentrycolor 1+
  db 51h
  db %1
%endmacro
; u16 color
%macro fillpalette 1+
  db 52h
  db %1
%endmacro
; [u16 color]
%macro setactivepalette 1+
  db 53h
  db %1
%endmacro
; [u16 color]
%macro fade 1+
  db 54h
  db %1
%endmacro
; u16 framesperfadeupdate
%macro setfadespeed 1+
  db 55h
  db %1
%endmacro
; 
%macro savepalette 0
  db 56h
%endmacro
; 
%macro restorepalette 0
  db 57h
%endmacro
; u8 displaypage
%macro setdisplaypage 1+
  db 58h
  db %1
%endmacro
; u8 drawpage
%macro setdrawpage 1+
  db 59h
  db %1
%endmacro
; string gp4file, u8 dispbytex, u16 dispy
%macro drawgp4 1+
  db 5ah
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax[, u8 paletteindex]
%macro fillrect 1+
  db 5bh
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax[, u8 paletteindex]
%macro fillrectxor 1+
  db 5ch
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 dstpageidx
%macro copypixels 1+
  db 5dh
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 dstpageidx
%macro swappixels 1+
  db 5eh
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 dstpageidx
%macro maskpixels 1+
  db 5fh
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 percolumndelay
%macro fadeinpixels 1+
  db 60h
  db %1
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 percolumndelay
%macro fadeoutpixels 1+
  db 61h
  db %1
%endmacro
;
%macro introanim 0
  db 62h
  db 00h
%endmacro
; u16 animidx, u16 seqidx, u16 s4dataptr
%macro animinit 1+
  db 63h
  db EXPTERM(EXPCONSTX(0))
  db %1
%endmacro
; u16 animidx
%macro animplay 1+
  db 63h
  db EXPTERM(EXPCONSTX(1))
  db %1
%endmacro
; u16 animidx
%macro animreqstop 1+
  db 63h
  db EXPTERM(EXPCONSTX(2))
  db %1
%endmacro
; u16 animidx
%macro animstop 1+
  db 63h
  db EXPTERM(EXPCONSTX(3))
  db %1
%endmacro
; u16 animidx
%macro animplayuntilend 1+
  db 63h
  db EXPTERM(EXPCONSTX(4))
  db %1
%endmacro
; u16 animidx, u16 destofs, u16 bufferofs
%macro animsetdatadisplacement 1+
  db 63h
  db EXPTERM(EXPCONSTX(5))
  db %1
%endmacro
;
%macro animreqstopall 0
  db 63h
  db EXPTERM(EXPCONSTX(6))
  db 00h
%endmacro
;
%macro animstopall 0
  db 63h
  db EXPTERM(EXPCONSTX(7))
  db 00h
%endmacro
; u16 animidx, u8 bufferidx
%macro animsetbufferidx 1+
  db 63h
  db EXPTERM(EXPCONSTX(8))
  db %1
%endmacro
; out u8 type
%macro getinteractiontype 1+
  db 64h
  db %1
%endmacro
; out u8 keystate
%macro getcontrolstate 1+
  db 65h
  db %1
%endmacro
; 
%macro clearscreen 0
  db 66h
%endmacro
; u16 frames
%macro wait 1+
  db 67h
  db %1
%endmacro
; out var year, out var month, out var day, out var hour, out var minute, out var second
%macro getdate 1+
  db 68h
  db %1
%endmacro
;
%macro timerstart 0
  db 69h
%endmacro
; out var framessincetimerstart
%macro timerend 1+
  db 6ah
  db %1
%endmacro
; 
%macro mouseshow 0
  db 6bh
  db EXPTERM(EXPCONSTX(0))
  db 00h
%endmacro
; 
%macro mousehide 0
  db 6bh
  db EXPTERM(EXPCONSTX(1))
  db 00h
%endmacro
; out u16 leftpressed, out u16 rightpressed
%macro mousegetbuttonstate 1+
  db 6bh
  db EXPTERM(EXPCONSTX(2))
  db %1
%endmacro
; out u16 xpos, out u16 ypos
%macro mousegetpos 1+
  db 6bh
  db EXPTERM(EXPCONSTX(3))
  db %1
%endmacro
; u16 xpos, u16 ypos
%macro mousesetpos 1+
  db 6bh
  db EXPTERM(EXPCONSTX(4))
  db %1
%endmacro
; u16 xmin, u16 xmax, u16 ymin, u16 ymax
%macro mousesetbounds 1+
  db 6bh
  db EXPTERM(EXPCONSTX(5))
  db %1
%endmacro
; u16 xmin, u16 xmax, u16 ymin, u16 ymax
%macro mousesetcursorimage 1+
  db 6bh
  db EXPTERM(EXPCONSTX(6))
  db %1
%endmacro
;
%macro mousestartblinking 0
  db 6bh
  db EXPTERM(EXPCONSTX(7))
  db 00h
%endmacro
;
%macro mousestopblinking 0
  db 6bh
  db EXPTERM(EXPCONSTX(8))
  db 00h
%endmacro
; u8 rate
%macro mousesetblindkrate 1+
  db 6bh
  db EXPTERM(EXPCONSTX(9))
  db %1
%endmacro
;  u8 diskindex, u8 driveindex
%macro promptdisk 1+
  db 6ch
  db %1
%endmacro
;  string mesfilename
%macro runmes 1+
  db 6dh
  db %1
%endmacro
;  string mllfilename
%macro loadmll 1+
  db 6eh
  db %1
%endmacro
;  string mesfilename
%macro callmes 1+
  db 6fh
  db %1
%endmacro
;
; 0x70 - illegal
;
; reg processortype
;  maybe 0 - 8086, 1 - 286, 2 - 386, 3 - 486, 4 - pentium
%macro getprocessortype 1+
  db 71h
  db EXPTERM(EXPCONSTX(0))
  db %1
  dw 0000h
%endmacro
; u16 rowoffset
%macro scroll 1+
  db 71h
  db EXPTERM(EXPCONSTX(1))
  db %1
  db 00h
%endmacro
; u16 xbytemin, u16 ymin, u16 xbytemax, u16 ymax, u8 srcpageidx, u16 xbytedst, u16 ydst, u8 dstpageidx
%macro dissolvein 1+
  db 71h
  db EXPTERM(EXPCONSTX(2))
  db %1
%endmacro
; u8 savenum, out destination, u16 length
%macro loadupdatesavepart 1+
  db 72h
  db %1
  db 00h
%endmacro
; u8 savenum, out destination, u16 length
%macro loadsavepart 1+
  db 73h
  db %1
  db 00h
%endmacro
; u8 savenum
%macro writesave 1+
  db 74h
  db %1
%endmacro
; u8 savenum
%macro loadsave 1+
  db 75h
  db %1
%endmacro
; u16 savenum, u16 memoffset
%macro writesavedata 1+
  db 76h
  db %1
  db 00h
%endmacro
; u16 savenum, u16 memoffset
%macro readsavedata 1+
  db 77h
  db %1
  db 00h
%endmacro
; string filename, u16 offset[, u8 filetype]
%macro readfile 1+
  db 78h
  db %1
%endmacro
; u16 savenum, out var year, out var month, out var day, out var hours, out var minutes, out var seconds
%macro getsavedate 1+
  db 79h
  db %1
%endmacro
; u16 dataoffset
%macro musicload 1+
  db 7ah
  db EXPTERM(EXPCONSTX(0))
  db %1
%endmacro
;
%macro musicplay 0
  db 7ah
  db EXPTERM(EXPCONSTX(1))
  db 00h
%endmacro
;
%macro musicstop 0
  db 7ah
  db EXPTERM(EXPCONSTX(2))
  db 00h
%endmacro
; u16 channel
%macro sfxchannel 1+
  db 7ah
  db EXPTERM(EXPCONSTX(3))
  db %1
%endmacro
; u16 dataoffset
%macro sfxload 1+
  db 7ah
  db EXPTERM(EXPCONSTX(4))
  db %1
%endmacro
;
%macro musicnop 0
  db 7ah
  db EXPTERM(EXPCONSTX(5))
%endmacro
; out var regionidx, u16 regionlistofs, u16 posx, u16 posy
%macro getfirstregioninside 1+
  db 7bh
  db %1
%endmacro
;
%macro refresh 0
  db 7ch
%endmacro
; u32 advnum
%macro mifes 0
  db 7dh
  db %1
%endmacro
; string filename, u16 offset, u16 length
%macro loadupdatesavepart2 1+
  db 7eh
  db %1
  db 00h
%endmacro
;
; 0x7f - illegal
;
;
%macro reopendiskpackage 0
  db 80h
%endmacro
;
; end of operators, higher level block macros follow
;
%assign meslastifnum 0
%assign mesifdepth 0
;
%macro if 1+
  %assign meslastifnum meslastifnum + 1
  ifop (meslastifnum & 0xffff), iflabel%[meslastifnum]_0, %1
  %assign mesifdepth mesifdepth + 1
  %xdefine mesfifdepthnum%[mesifdepth] meslastifnum
  %assign mesfifelse%[mesifdepth] 0
%endmacro
;
%macro else 0
  iflabel%[mesfifdepthnum%[mesifdepth]]_%[mesfifelse%[mesifdepth]]:
  %assign mesfifelse%[mesifdepth] mesfifelse%[mesifdepth] + 1
  elseop mesfifdepthnum%[mesifdepth], iflabel%[mesfifdepthnum%[mesifdepth]]_%[mesfifelse%[mesifdepth]]
%endmacro
;
%macro endif 0
  iflabel%[mesfifdepthnum%[mesifdepth]]_%[mesfifelse%[mesifdepth]]:
  %assign mesifdepth mesifdepth - 1
%endmacro
;
%assign meslastwhilenum 0
%assign meswhiledepth 0
;
%macro while 1+
  %assign meslastwhilenum meslastwhilenum + 1
  startloopop whilelabel%[meslastwhilenum], %1
  %assign meswhiledepth meswhiledepth + 1
  %xdefine mesfwhiledepthnum%[meswhiledepth] meslastwhilenum
%endmacro
;
%macro wend 0
  continue
  whilelabel%[mesfwhiledepthnum%[meswhiledepth]]:
  break
  %assign meswhiledepth meswhiledepth - 1
%endmacro
;
%assign meslastmenunum 0
;
%macro menu 1+
  %assign meslastmenunum meslastmenunum + 1
  startmenuop mesmenulabel%[meslastmenunum], %1
%endmacro
;
%macro endmenu 0
  endofcode
  mesmenulabel%[meslastmenunum]:
%endmacro
;
%macro choice 1+
  startmenuitemop %1
%endmacro
;
%macro endchoice 0
  endmenuitemop
  endofcode
%endmacro
;
%assign meslastselectnum 0
%assign messelectdepth 0
;
%macro select 1+
  %assign meslastselectnum meslastselectnum + 1
  selectop messelectlabel_%[meslastselectnum], %1
  %assign messelectdepth messelectdepth + 1
  %xdefine messelectdepthnum%[messelectdepth] meslastselectnum
  %assign messelectcasenum%[messelectdepth] 0
%endmacro
;
%macro endselect 0
  messelectlabel_%[messelectdepthnum%[messelectdepth]]:
  %assign messelectdepth messelectdepth - 1
  break
%endmacro
;
%macro case 1+
  %assign messelectcasenum%[messelectdepth] messelectcasenum%[messelectdepth] + 1
  caseop mescaselabel%[messelectdepthnum%[messelectdepth]]_%[messelectcasenum%[messelectdepth]], %1
%endmacro
;
%macro endcase 0
  mescaselabel%[messelectdepthnum%[messelectdepth]]_%[messelectcasenum%[messelectdepth]]:
%endmacro
;
; build macros
;
; 0 - get make dependencies
; 1 - get global addresses (map)
; 2 - assemble binary
%ifndef EXTERNPASS
  %define EXTERNPASS 2
%endif
;
%macro pdep 1
  %if EXTERNPASS == 0
    %depend %1
  %elif EXTERNPASS == 2
    %include %1
  %endif
%endmacro
;
%macro pextern 1+
  %if EXTERNPASS < 2
    %rep %0
      %define %1 0
      %rotate 1
    %endrep
  %endif
%endmacro
;
%ifndef MESOFFSET
  %define MESOFFSET 0
%endif
;
org MESOFFSET
cpu 186
;
%endif