#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "sjisdec2.h"
#include "sjismap.h"

unsigned int writesjisbytes(char const* src, unsigned int srclen, char* buffer, unsigned int buffered, FILE* effout) {
  for (unsigned int i = 0; i < srclen; ++i) {
    unsigned int cc = (unsigned int)src[i] & 0xff;
    unsigned int plen = 3 + (i != 0) * 2 + (cc >= 0xa0);
    if (buffered + plen > 2048) { // bufferlen
      if (buffered != fwrite(buffer, 1, buffered, effout)) {
        // err
      }
      buffered = 0;
    }
    buffer[buffered] = ',';
    buffer[buffered + 1] = ' ';
    buffered += (i != 0) << 1;
    buffer[buffered] = '0';
    buffered += (cc >= 0xa0);
    unsigned char temp = cc >> 4 & 0x0f;
    buffer[buffered] = '0' + temp + ((9 - temp) >> 4 & 0x27);
    temp = cc & 0x0f;
    buffer[buffered + 1] = '0' + temp + ((9 - temp) >> 4 & 0x27);
    buffer[buffered + 2] = 'h';
    buffered += 3;
  }
  return buffered;
}

unsigned int writebytes(char const* src, unsigned int srclen, char* buffer, unsigned int buffered, FILE* outfile) {
  while (srclen > 0) {
    unsigned int maxcopy = buffered + srclen > 2048 ? 2048 - buffered : srclen;
    if (maxcopy) {
      memcpy(buffer + buffered, src, maxcopy);
      srclen -= maxcopy;
      src += maxcopy;
    }
    buffered += maxcopy;
    if (buffered >= 2048) {
      fwrite(buffer, 1, buffered, outfile);
      buffered = 0;
    }
  }
  return buffered;
}

unsigned int matchstr(char const* str, unsigned char const* retryofs, unsigned int matchlen, char c) {
  unsigned char matched;
  while (!(matched = (c == str[matchlen])) && matchlen) {
    matchlen = retryofs[matchlen];
  }
  return matchlen + matched;
}

unsigned int counthexchars(char const* str, unsigned int slen, unsigned int max) {
  unsigned int mlen = 0;
  unsigned int max2 = slen < max ? slen : max;
  char c;
  while (mlen < max2) {
    c = str[mlen];
    if (((c >= '0') & (c <= '9')) | ((c >= 'a') & (c <= 'f')) | ((c >= 'A') & (c <= 'F'))) {
      ++mlen;
    } else {
      max2 = mlen;
    }
  }
  return mlen;
}

int fromhex(char const* str, int count) {
  int val = 0;
  for (int i = 0; i < count; ++i) {
    char c = str[i];
    val = val << 4 | (c >= 'A' ? 9 + (c & 0x1f) : c - '0');
  }
  return val;
}

unsigned int convertutf8tosjis2escaped(char const* u8src, unsigned int srclen, char* sjisdst, unsigned short const* values, short const* lookup) {
  unsigned int declen = 0;
  unsigned int start = 0;
  unsigned int end = 0;
  while (end < srclen) {
    unsigned int countch = 0;
    if (u8src[end] == '{') {
      if (end > start) {
        declen += convertutf8tosjis2(u8src + start, end - start, sjisdst + declen, values, lookup);
      }
      countch = counthexchars(u8src + end + 1, srclen - end - 1, 4);
      if ((countch == 2 || countch == 4) && (end + 1 + countch < srclen && u8src[end + 1 + countch] == '}')) {
        int v = fromhex(u8src + end + 1, countch);
        sjisdst[declen] = (char)(v >> 8 & 0xff);
        sjisdst[declen + 1] = (char)(v & 0xff);
        declen += 2;
        start = end + 2 + countch;
        end = start;
      } else {
        start = end + 1;
        end = (end + 1 < srclen ? end + 2 : start);
      }
    } else {
      ++end;
    }
  }
  if (end > start) {
    declen += convertutf8tosjis2(u8src + start, end - start, sjisdst + declen, values, lookup);
  }
  return declen;
}

unsigned char makelookup(FILE* effin, int* lookupnum, unsigned short* sjislookup, unsigned char* indexlookup, int maxlookup) {
  unsigned char success = 1;
  char temp[2048];
  char enctemp[2048];
  char dectemp[2048];
  //char outtemp[2048];
  unsigned int encodedlen = 0;
  //unsigned int outlen = 0;
  unsigned int readamt = 2048;
  unsigned char decoding = 0;
  char const matching1[] = "$CJIS(";
  unsigned char retryofs1[] = {0, 0, 0, 0, 0, 0};

  unsigned int startmatchlen = strlen(matching1);

  unsigned int startinglen = 0;
  unsigned int closinglen = 0;

  int singlecounts[256];
  int* doublecounts = malloc(sizeof(int) * 32768);
  unsigned short* sorting = doublecounts != NULL ? malloc(sizeof(unsigned short) * 10000) : NULL;

  if (sorting == NULL) {
    free(doublecounts);
    printf("Could not allocate memory while making lookup table.\n");
    fflush(stdout);
    return 0;
  }

  memset(singlecounts, 0, sizeof(int) * 256);
  memset(doublecounts, 0, sizeof(int) * 32768);

  unsigned int matchlen = 0;
  while (readamt == 2048 && (readamt = fread(temp + matchlen, 1, 2048 - matchlen, effin))) {
    readamt += matchlen;
    unsigned int lastidx = 0;
    for (unsigned int i = matchlen; success && i < readamt; ++i) {
      if (decoding) {
        if (closinglen == startinglen) {
          if (temp[i] == '"') {
            --closinglen;
          }
        } else {
          if (temp[i] == ')') {
            if (--closinglen == 0) {
              unsigned int addlen = i + 1 - matchlen - lastidx;
              if (encodedlen + addlen <= 2048) {
                memcpy(enctemp + encodedlen, temp + lastidx, addlen);
                encodedlen += addlen;
                encodedlen -= startinglen;
                unsigned int decodedlen = convertutf8tosjis2escaped(enctemp, encodedlen, dectemp, cvtsjisvalues, cvtsjislookuptable);
                encodedlen = 0;

                int numdouble = 0;
                for (unsigned int idx = 0; idx + 2 <= decodedlen; idx += 2) {
                  numdouble += (dectemp[idx] != 0);
                }

                if (numdouble > 0) {
                  for (unsigned int idx = 0; idx + 2 <= decodedlen; idx += 2) {
                    int v = ((int)dectemp[idx] & 0xff) << 8 | ((int)dectemp[idx + 1] & 0xff);
                    if (v >= 0x8000) {
                      ++doublecounts[v - 0x8000];
                    } else if (v < 256) {
                      ++singlecounts[v];
                    }
                  }
                }

                //outlen = writesjisbytes(dectemp, decodedlen, outtemp, outlen, effout);
                lastidx = i + 1;
              } else {
                printf("CJIS has too many characters to decode\n");
                fflush(stdout);
                success = 0;
              }
              decoding = 0;
              matchlen = 0;
            }
          } else {
            closinglen = startinglen - (temp[i] == '"');
          }
        }
      } else { // not decoding (yet)
        if (matchlen < startmatchlen) {
          matchlen = matchstr(matching1, retryofs1, matchlen, temp[i]);
        } else {
          if (temp[i] == '"') {
            //unsigned int addlen = i - matchlen - lastidx;
            //outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
            startinglen = matchlen - startmatchlen + 2;
            decoding = 1;
            lastidx = i + 1;
            matchlen = 0;
            closinglen = startinglen;
          } else if (temp[i] == '(') {
            ++matchlen;
          } else {
            matchlen = 0;
          }
        }
      }
    }
    if (lastidx + matchlen < readamt) {
      unsigned int addlen = readamt - matchlen - lastidx;
      if (decoding) {
        if (encodedlen + addlen > 2048) {
          printf("CJIS has too many characters to decode\n");
          fflush(stdout);
          success = 0;
        } else {
          memcpy(enctemp + encodedlen, temp + lastidx, addlen);
          encodedlen += addlen;
        }
      } else {
        //outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
      }
    }
    if (matchlen) {
      for (unsigned int i = 0; i < matchlen; ++i) {
        temp[i] = temp[2048 - matchlen + i];
      }
    }
  }
  //if (outlen) {
  //  fwrite(outtemp, 1, outlen, effout);
  //}
  if (decoding) {
    printf("Unclosed CJIS block\n");
    fflush(stdout);
    success = 0;
  }
  if (success) {
    for (int idx = 0; idx < 32; ++idx) {
      singlecounts[idx] = 0;
    }
    for (int idx = 128; idx < 160; ++idx) {
      singlecounts[idx] = 0;
    }
    for (int idx = 192; idx < 256; ++idx) {
      singlecounts[idx] = 0;
    }
    int numsingle = 0;
    for (int idx = 32; idx < 192; ++idx) {
      if (singlecounts[idx] != 0) {
        singlecounts[numsingle] = idx;
        ++numsingle;
      }
    }
    int numdouble = 0;
    for (int idx = 0; idx < 32768; ++idx) {
      if (doublecounts[idx] > 2) {
        sorting[numdouble] = idx;
        ++numdouble;
      }
    }
    int sortidx = 0;
    while (sortidx + 1 < numdouble) {
      if (doublecounts[sorting[sortidx]] < doublecounts[sorting[sortidx + 1]]) {
        unsigned short temp = sorting[sortidx];
        sorting[sortidx] = sorting[sortidx + 1];
        sorting[sortidx + 1] = temp;
        sortidx -= (sortidx != 0);
      } else {
        ++sortidx;
      }
    }

    int lnum = numsingle + numdouble;
    lnum = lnum > maxlookup ? maxlookup : lnum;
    *lookupnum = lnum;
    for (int idx = 0; idx < numsingle; ++idx) {
      sjislookup[idx] = singlecounts[idx];
    }
    for (int idx = numsingle; idx < lnum; ++idx) {
      sjislookup[idx] = sorting[idx - numsingle] + 0x8000;
    }
    for (int idx = 0; idx < lnum; ++idx) {
      indexlookup[idx] = (unsigned char)idx;
    }
    sortidx = 0;
    while (sortidx + 1 < lnum) {
      if (sjislookup[indexlookup[sortidx]] > sjislookup[indexlookup[sortidx + 1]]) {
        unsigned char temp = indexlookup[sortidx];
        indexlookup[sortidx] = indexlookup[sortidx + 1];
        indexlookup[sortidx + 1] = temp;
        sortidx -= (sortidx != 0);
      } else {
        ++sortidx;
      }
    }
  }
  free(sorting);
  free(doublecounts);
  return success;
}

unsigned char convertfile(FILE* effin, FILE* effout, int lookupnum, unsigned short const* sjislookup, unsigned char const* indexlookup, char haslookup) {
  unsigned char success = 1;
  char temp[2048];
  char enctemp[2048];
  char dectemp[2048];
  char outtemp[2048];
  char cdectemp[2049];
  unsigned int encodedlen = 0;
  unsigned int outlen = 0;
  unsigned int readamt = 2048;
  unsigned char decoding = 0;
  char const matching1[] = "$CJIS(";
  unsigned char retryofs1[] = {0, 0, 0, 0, 0, 0};

  unsigned int startmatchlen = strlen(matching1);

  unsigned int startinglen = 0;
  unsigned int closinglen = 0;

  char tabletemp[168 * 2];
  if (!haslookup || lookupnum == 0) {
    sprintf(tabletemp, "dw %d\n\n", 2);
  } else {
    sprintf(tabletemp, "dw %d\ndb ", (lookupnum + 1) * 2);
  }
  outlen = writebytes(tabletemp, strlen(tabletemp), outtemp, outlen, effout);

  if (haslookup) {
    if (lookupnum > 0) {
      for (int idx = 0; idx < lookupnum; ++idx) {
        int v = sjislookup[idx];
        tabletemp[idx * 2] = (v >> 8 & 0xff);
        tabletemp[idx * 2 + 1] = (v & 0xff);
      }
      outlen = writesjisbytes(tabletemp, lookupnum * 2, outtemp, outlen, effout);
      tabletemp[0] = '\n';
      tabletemp[1] = '\n';
      outlen = writebytes(tabletemp, 2, outtemp, outlen, effout);
    }
  }

  unsigned int matchlen = 0;
  while (readamt == 2048 && (readamt = fread(temp + matchlen, 1, 2048 - matchlen, effin))) {
    readamt += matchlen;
    unsigned int lastidx = 0;
    for (unsigned int i = matchlen; success && i < readamt; ++i) {
      if (decoding) {
        if (closinglen == startinglen) {
          if (temp[i] == '"') {
            --closinglen;
          }
        } else {
          if (temp[i] == ')') {
            if (--closinglen == 0) {
              unsigned int addlen = i + 1 - matchlen - lastidx;
              if (encodedlen + addlen <= 2048) {
                memcpy(enctemp + encodedlen, temp + lastidx, addlen);
                encodedlen += addlen;
                encodedlen -= startinglen;
                unsigned int decodedlen = convertutf8tosjis2escaped(enctemp, encodedlen, dectemp, cvtsjisvalues, cvtsjislookuptable);
                encodedlen = 0;
                int doublecount = 0;
                for (unsigned int idx = 0; idx < decodedlen; idx += 2) {
                  doublecount += (dectemp[idx] != 0);
                }
                int d2len = 1;
                if (doublecount > 0) {
                  cdectemp[0] = 1;
                  for (unsigned int idx = 0; idx < decodedlen; idx += 2) {
                    unsigned short v = ((unsigned short)dectemp[idx] & 0xff) << 8 | ((unsigned short)dectemp[idx + 1] & 0xff);
                    if (v >= 32) {
                      int schidx = 0;
                      int schlen = lookupnum;
                      while (schlen > 1) {
                        int halflen = schlen >> 1;
                        if (sjislookup[indexlookup[schidx + halflen]] <= v) {
                          schlen -= halflen;
                          schidx += halflen;
                        } else {
                          schlen = halflen;
                        }
                      }
                      if (schlen == 1 && sjislookup[indexlookup[schidx]] == v) {
                        int lkidx = indexlookup[schidx] + 1;
                        lkidx = (lkidx >= 0x69) ? lkidx - 0x69 + 0xa0 : lkidx - 0x01 + 0x18;
                        cdectemp[d2len++] = (char)lkidx;
                      } else {
                        cdectemp[d2len] = dectemp[idx];
                        cdectemp[d2len + 1] = dectemp[idx + 1];
                        d2len += 2;
                      }
                    } else {
                      if (v == '\n') {
                        cdectemp[d2len++] = 4;
                      } else if (v == '\r') {
                        if (idx + 2 < decodedlen && (dectemp[idx + 2] == 0) && (dectemp[idx + 3] == '\n')) {
                          idx += 2;
                        }
                        cdectemp[d2len++] = 4;
                      }
                    }
                  }
                } else {
                  cdectemp[0] = 2;
                  for (unsigned int idx = 0; idx < decodedlen; idx += 2) {
                    char c = dectemp[idx + 1];
                    if (c == '\n') {
                      cdectemp[d2len++] = 4;
                    } else if (c == '\r') {
                      if (idx + 2 < decodedlen && (dectemp[idx + 2] == 0) && (dectemp[idx + 3] == '\n')) {
                        idx += 2;
                      }
                      cdectemp[d2len++] = 4;
                    } else {
                      cdectemp[d2len++] = c;
                    }
                  }
                }
                outlen = writesjisbytes(cdectemp, d2len, outtemp, outlen, effout);
                lastidx = i + 1;
              } else {
                printf("CJIS has too many characters to decode\n");
                fflush(stdout);
                success = 0;
              }
              decoding = 0;
              matchlen = 0;
            }
          } else {
            closinglen = startinglen - (temp[i] == '"');
          }
        }
      } else { // not decoding (yet)
        if (matchlen < startmatchlen) {
          matchlen = matchstr(matching1, retryofs1, matchlen, temp[i]);
        } else {
          if (temp[i] == '"') {
            unsigned int addlen = i - matchlen - lastidx;
            outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
            startinglen = matchlen - startmatchlen + 2;
            decoding = 1;
            lastidx = i + 1;
            matchlen = 0;
            closinglen = startinglen;
          } else if (temp[i] == '(') {
            ++matchlen;
          } else {
            matchlen = 0;
          }
        }
      }
    }
    if (lastidx + matchlen < readamt) {
      unsigned int addlen = readamt - matchlen - lastidx;
      if (decoding) {
        if (encodedlen + addlen > 2048) {
          printf("CJIS has too many characters to decode\n");
          fflush(stdout);
          success = 0;
        } else {
          memcpy(enctemp + encodedlen, temp + lastidx, addlen);
          encodedlen += addlen;
        }
      } else {
        outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
      }
    }
    if (matchlen) {
      for (unsigned int i = 0; i < matchlen; ++i) {
        temp[i] = temp[2048 - matchlen + i];
      }
    }
  }
  if (outlen) {
    fwrite(outtemp, 1, outlen, effout);
  }
  if (decoding) {
    printf("Unclosed CJIS block\n");
    fflush(stdout);
    success = 0;
  }
  return success;
}

int parsenum(char const* str) {
  int len = strlen(str);
  char isdec = (len > 0);
  int val = 0;
  for (int i = 0; i < len; ++i) {
    char c = str[i];
    isdec &= (c >= '0' && c <= '9');
    val = val * 10 + (c - '0');
  }
  return isdec ? val : -1;
}

int main(int argc, char const** argv) {
  int const maxlookup = 168;
  int lookupnum = 1;
  unsigned short sjislookup[maxlookup];
  unsigned char indexlookup[maxlookup];
  indexlookup[0] = 0;
  sjislookup[0] = 0;

  unsigned char success = 0;
  if (argc == 3 || argc == 4) {
    FILE* effin = fopen(argv[1], "rb");
    if (effin) {
      FILE* effout = fopen(argv[2], "wb");
      if (effout) {
        char haslookup = !(argc == 4 && parsenum(argv[3]) == 1);
        success = 1;
        if (haslookup) {
          success = makelookup(effin, &lookupnum, sjislookup, indexlookup, maxlookup);
          if (success) {
            success = (fseek(effin, 0, SEEK_SET) == 0);
          }
        }
        if (success) {
          success = convertfile(effin, effout, lookupnum, sjislookup, indexlookup, haslookup);
        }
        fclose(effout);
      } else {
        printf("FERMSJIS couldn't open %s for writing\n", argv[2]);
        fflush(stdout);
      }
      fclose(effin);
    } else {
      printf("FERMSJIS couldn't open %s for reading\n", argv[1]);
      fflush(stdout);
    }
  } else {
    printf("FERMSJIS wrong number of parameters\n");
  }
  return success != 0 ? 0 : 1;
}