#include "sjisdec2.h"

unsigned int convertutf8tosjis2(char const* u8src, unsigned int srclen, char* sjisdst, unsigned short const* values, short const* lookup) {
  unsigned int resultlen = 0;
  for (unsigned int i = 0; i < srclen; ++i) {
    short bidx = 0;
    unsigned int relofs = 0;
    unsigned short fallbackidx = 0;
    unsigned int fallbackrel = 0;
    do {
      bidx = lookup[((unsigned int)-bidx) * 256 + ((unsigned int)u8src[i + relofs++] & 0xff)];
      if (bidx < 0 && lookup[((unsigned int)-bidx) * 256] > 0) {
        fallbackidx = lookup[((unsigned int)-bidx) * 256] - 1;
        fallbackrel = relofs;
      }
    } while (bidx < 0 && i + relofs < srclen);
    if (bidx > 0) { // defined
      unsigned short val = values[bidx - 1];
      sjisdst[resultlen] = (char)(val >> 8 & 0xff);
      sjisdst[resultlen + 1] = (char)(val & 0xff);
      resultlen += 2;
      i += relofs - 1;
    } else if (bidx == 0) { // undefined
      if (fallbackrel) {
        unsigned short val = values[fallbackidx];
        sjisdst[resultlen] = (char)(val >> 8 & 0xff);
        sjisdst[resultlen + 1] = (char)(val & 0xff);
        resultlen += 2;
        i += fallbackrel - 1;
      } else { // output a single byte and continue
        sjisdst[resultlen] = (char)0;
        sjisdst[resultlen + 1] = u8src[i];
        resultlen += 2;
      }
    } else {
      if (fallbackrel) {
        unsigned short val = values[fallbackidx];
        sjisdst[resultlen] = (char)(val >> 8 & 0xff);
        sjisdst[resultlen + 1] = (char)(val & 0xff);
        resultlen += 2;
        i += fallbackrel - 1;
      } else { // output a single byte and continue
        sjisdst[resultlen] = (char)0;
        sjisdst[resultlen + 1] = u8src[i];
        resultlen += 2;
      }
    }
  }
  return resultlen;
}