// convert nasm symbol map on stdin to asm file for inclusion on stdout

#include <stdio.h>
#include <string.h>

char nextline(FILE* file, char* buffer, unsigned int buffersize, int* buffered, int* lineofs, int* lineend, char* state) {
  int nstate = *state;
  char eos = 0;
  int blen = *buffered;
  int ofs = *lineend;
  if ((nstate & 1)) {
    if (ofs < blen) {
      ofs += (buffer[ofs] == '\n');
    } else {
      ofs = 0;
      blen = fread(buffer, 1, buffersize, file);
      *buffered = blen;
      if (ofs < blen) {
        ofs += (buffer[ofs] == '\n');
      }
    }
  }
  nstate = 0;
  int end = ofs;
  char foundend = 0;
  do {
    if (end == blen) {
      if ((unsigned int)blen == buffersize) {
        if (ofs > 0) {
          memmove(buffer, buffer + ofs, end - ofs);
          end -= ofs;
          blen -= ofs;
          ofs = 0;
        }
        if ((unsigned int)blen < buffersize) {
          blen += fread(buffer + blen, 1, buffersize - blen, file);
          *buffered = blen;
        }
      } else {
        eos = 1;
      }
    }
    foundend = 1;
    while (end < blen && !(foundend = ((buffer[end] == '\n') | (buffer[end] == '\r')))) {
      ++end;
    }
    if (end < blen) {
      nstate = (buffer[end] == '\r');
      ++end;
    }
  } while (!foundend);
  *state = nstate;
  *lineofs = ofs;
  *lineend = end;
  return !eos;
}

int getnonspacetokens(char const* str, int ofs, int end, int* ofspairs, int maxtokens) {
  int numtokens = 0;
  while (ofs < end && numtokens < maxtokens) {
    while (ofs < end && str[ofs] == ' ') {
      ++ofs;
    }
    int startofs = ofs;
    while (ofs < end && str[ofs] != ' ') {
      ++ofs;
    }
    int endofs = ofs;
    if (startofs < endofs) {
      ofspairs[numtokens * 2] = startofs;
      ofspairs[numtokens * 2 + 1] = endofs;
      ++numtokens;
    }
  }
  return numtokens;
}

void bufferwrite(FILE* file, char* writebuffer, int buffersize, char const* data, int len, int* buffered) {
  int blen = *buffered;
  if (blen + len <= buffersize) {
    memcpy(writebuffer + blen, data, len);
    *buffered = blen + len;
  } else {
    while (len > 0) {
      int rem = buffersize - blen;
      int clen = len > rem ? rem : len;
      memcpy(writebuffer + blen, data, clen);
      blen += clen;
      fwrite(writebuffer, 1, blen, file);
      blen = 0; //
      data += clen;
      len -= clen;
    }
    *buffered = blen;
  }
}

int main(int argc, char const** argv) {
  if (argc >= 1) {
    int preflen = argc >= 2 ? strlen(argv[1]) : 0;
    int const readbuffersize = 16384;
    char readbuffer[readbuffersize];
    int const writebuffersize = 16384;
    char writebuffer[writebuffersize];
    int writebuffered = 0;
    int readbuffered = 0;
    int lineofs = 0;
    int lineend = 0;
    char state = 0;
    char hasmore = 1;
    char symbols = 0;
    int numdefs = 0;
    int ofspairs[6];
    readbuffered = fread(readbuffer, 1, readbuffersize, stdin);
    do {
      hasmore = nextline(stdin, readbuffer, readbuffersize, &readbuffered, &lineofs, &lineend, &state);
      int len2 = lineend - lineofs;
      if (len2 >= 1 && ((readbuffer[lineend - 1] == '\r') | (readbuffer[lineend - 1] == '\n'))) {
        --len2;
      }
      if (len2 >= 2 && ((readbuffer[lineend - 2] == '\r'))) {
        --len2;
      }
      if (symbols) {
        if (len2 == 0) {
          symbols = 0;
        } else {
          int numtok = getnonspacetokens(readbuffer, lineofs, lineofs + len2, ofspairs, 3);
          if (numtok == 3) {
            int tlen = ofspairs[2 * 2 + 1] - ofspairs[2 * 2];
            if (tlen >= preflen && (preflen == 0 || memcmp(argv[1], readbuffer + ofspairs[2 * 2], preflen) == 0)) {
              if (numdefs > 0) {
                bufferwrite(stdout, writebuffer, writebuffersize, "\n", 1, &writebuffered);
              }
              bufferwrite(stdout, writebuffer, writebuffersize, "%define ", 8, &writebuffered);
              bufferwrite(stdout, writebuffer, writebuffersize, readbuffer + ofspairs[2 * 2], tlen, &writebuffered);
              bufferwrite(stdout, writebuffer, writebuffersize, " ", 1, &writebuffered);
              bufferwrite(stdout, writebuffer, writebuffersize, readbuffer + ofspairs[1 * 2], ofspairs[1 * 2 + 1] - ofspairs[1 * 2], &writebuffered);
              bufferwrite(stdout, writebuffer, writebuffersize, "H", 1, &writebuffered);
              ++numdefs;
            }
          }
        }
      }
      if (len2 >= 5 && memcmp(readbuffer + lineofs, "Real ", 5) == 0) {
        symbols = 1;
      }
    } while (hasmore);
    if (writebuffered > 0) {
      fwrite(writebuffer, 1, writebuffered, stdout);
    }
  }
  return 0;
}