#include <stdio.h>
#include <string.h>

#include "sjisdec.h"
#include "sjismap.h"

unsigned int writesjisbytes(char const* src, unsigned int srclen, char* buffer, unsigned int buffered, FILE* effout) {
  for (unsigned int i = 0; i < srclen; ++i) {
    unsigned int cc = (unsigned int)src[i] & 0xff;
    unsigned int plen = 3 + (i != 0) * 2 + (cc >= 0xa0);
    if (buffered + plen > 2048) { // bufferlen
      if (buffered != fwrite(buffer, 1, buffered, effout)) {
        // err
      }
      buffered = 0;
    }
    buffer[buffered] = ',';
    buffer[buffered + 1] = ' ';
    buffered += (i != 0) << 1;
    buffer[buffered] = '0';
    buffered += (cc >= 0xa0);
    unsigned char temp = cc >> 4 & 0x0f;
    buffer[buffered] = '0' + temp + ((9 - temp) >> 4 & 0x27);
    temp = cc & 0x0f;
    buffer[buffered + 1] = '0' + temp + ((9 - temp) >> 4 & 0x27);
    buffer[buffered + 2] = 'h';
    buffered += 3;
  }
  return buffered;
}

unsigned int writebytes(char const* src, unsigned int srclen, char* buffer, unsigned int buffered, FILE* outfile) {
  while (srclen > 0) {
    unsigned int maxcopy = buffered + srclen > 2048 ? 2048 - buffered : srclen;
    if (maxcopy) {
      memcpy(buffer + buffered, src, maxcopy);
      srclen -= maxcopy;
      src += maxcopy;
    }
    buffered += maxcopy;
    if (buffered >= 2048) {
      fwrite(buffer, 1, buffered, outfile);
      buffered = 0;
    }
  }
  return buffered;
}

unsigned int matchstr(char const* str, unsigned char const* retryofs, unsigned int matchlen, char c) {
  unsigned char matched;
  while (!(matched = (c == str[matchlen])) && matchlen) {
    matchlen = retryofs[matchlen];
  }
  return matchlen + matched;
}

unsigned char convertfile(FILE* effin, FILE* effout) {
  unsigned char success = 1;
  char temp[2048];
  char enctemp[2048];
  char dectemp[2048];
  char outtemp[2048];
  unsigned int encodedlen = 0;
  unsigned int outlen = 0;
  unsigned int readamt = 2048;
  unsigned char decoding = 0;
  char const matching1[] = "$SJIS(";
  unsigned char retryofs1[] = {0, 0, 0, 0, 0, 0};

  unsigned int startmatchlen = strlen(matching1);

  unsigned int startinglen = 0;
  unsigned int closinglen = 0;

  unsigned int matchlen = 0;
  while (readamt == 2048 && (readamt = fread(temp + matchlen, 1, 2048 - matchlen, effin))) {
    readamt += matchlen;
    unsigned int lastidx = 0;
    for (unsigned int i = matchlen; success && i < readamt; ++i) {
      if (decoding) {
        if (closinglen == startinglen) {
          if (temp[i] == '"') {
            --closinglen;
          }
        } else {
          if (temp[i] == ')') {
            if (--closinglen == 0) {
              unsigned int addlen = i + 1 - matchlen - lastidx;
              if (encodedlen + addlen <= 2048) {
                memcpy(enctemp + encodedlen, temp + lastidx, addlen);
                encodedlen += addlen;
                encodedlen -= startinglen;
                unsigned int decodedlen = convertutf8tosjis(enctemp, encodedlen, dectemp, cvtsjisvalues, cvtsjislookuptable);
                encodedlen = 0;
                outlen = writesjisbytes(dectemp, decodedlen, outtemp, outlen, effout);
                lastidx = i + 1;
              } else {
                printf("SJIS has too many characters to decode\n");
                fflush(stdout);
                success = 0;
              }
              decoding = 0;
              matchlen = 0;
            }
          } else {
            closinglen = startinglen - (temp[i] == '"');
          }
        }
      } else { // not decoding (yet)
        if (matchlen < startmatchlen) {
          matchlen = matchstr(matching1, retryofs1, matchlen, temp[i]);
        } else {
          if (temp[i] == '"') {
            unsigned int addlen = i - matchlen - lastidx;
            outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
            startinglen = matchlen - startmatchlen + 2;
            decoding = 1;
            lastidx = i + 1;
            matchlen = 0;
            closinglen = startinglen;
          } else if (temp[i] == '(') {
            ++matchlen;
          } else {
            matchlen = 0;
          }
        }
      }
    }
    if (lastidx + matchlen < readamt) {
      unsigned int addlen = readamt - matchlen - lastidx;
      if (decoding) {
        if (encodedlen + addlen > 2048) {
          printf("SJIS has too many characters to decode\n");
          fflush(stdout);
          success = 0;
        } else {
          memcpy(enctemp + encodedlen, temp + lastidx, addlen);
          encodedlen += addlen;
        }
      } else {
        outlen = writebytes(temp + lastidx, addlen, outtemp, outlen, effout);
      }
    }
    if (matchlen) {
      for (unsigned int i = 0; i < matchlen; ++i) {
        temp[i] = temp[2048 - matchlen + i];
      }
    }
  }
  if (outlen) {
    fwrite(outtemp, 1, outlen, effout);
  }
  if (decoding) {
    printf("Unclosed SJIS block\n");
    fflush(stdout);
    success = 0;
  }
  return success;
}

int main(int argc, char const** argv) {
  unsigned char success = 0;
  if (argc == 3) {
    FILE* effin = fopen(argv[1], "rb");
    if (effin) {
      FILE* effout = fopen(argv[2], "wb");
      if (effout) {
        success = convertfile(effin, effout);
        fclose(effout);
      } else {
        printf("TOSJIS couldn't open %s for writing\n", argv[2]);
        fflush(stdout);
      }
      fclose(effin);
    } else {
      printf("TOSJIS couldn't open %s for reading\n", argv[1]);
      fflush(stdout);
    }
  } else {
    printf("TOSJIS wrong number of parameters\n");
  }
  return success != 0 ? 0 : 1;
}